#include "psp_pdf.h"

fz_renderer  *drawgc    = nil;
fz_error     *error     = nil;
pdf_xref     *src       = nil;
pdf_pagetree *srcpages  = nil;
char         cachesrv_status;
TABCACHE     tab_cache[3];

void zoom_page(int cache_indx)
{
  fz_matrix ctm;
  fz_irect bbox;
  int w;
  int h;
  int bh;
  int x;
  int y;
  int drawrotate         = 0;
  int drawbands          = 1;
  fz_error *error;
  unsigned char *ptr;

  ctm = fz_identity();
  ctm = fz_concat(ctm, fz_translate(0, -tab_cache[cache_indx].drawpage->mediabox.y1));
  ctm = fz_concat(ctm, fz_scale(tab_cache[cache_indx].zoom, -tab_cache[cache_indx].zoom));
  ctm = fz_concat(ctm, fz_rotate(drawrotate + tab_cache[cache_indx].drawpage->rotate));

 

   bbox = fz_roundrect(fz_transformaabb(ctm, tab_cache[cache_indx].drawpage->mediabox));
   w = bbox.x1 - bbox.x0;
   h = bbox.y1 - bbox.y0;
   bh = h / drawbands;

   error = fz_newpixmap(&tab_cache[cache_indx].pix, bbox.x0, bbox.y0, w, bh, 4);
   if (error)
     die(error);

   memset(tab_cache[cache_indx].pix->samples, 0xff, tab_cache[cache_indx].pix->h * tab_cache[cache_indx].pix->w * tab_cache[cache_indx].pix->n);
 
   error = fz_rendertreeover(drawgc, tab_cache[cache_indx].pix, tab_cache[cache_indx].drawpage->tree, ctm);
   if (error)
    { 
	maybedie(error);
    }
//
//             Sumatra library use ARGB buffers psp use RGBA so we transform it
//	
    ptr = tab_cache[cache_indx].pix->samples;

    for(y=0; y<tab_cache[cache_indx].pix->h ;y++)
      for(x=0; x <tab_cache[cache_indx].pix->w; x++)
       {
         int a = tab_cache[cache_indx].pix->samples[x * tab_cache[cache_indx].pix->n + y * tab_cache[cache_indx].pix->w * tab_cache[cache_indx].pix->n + 0];
         int r = tab_cache[cache_indx].pix->samples[x * tab_cache[cache_indx].pix->n + y * tab_cache[cache_indx].pix->w * tab_cache[cache_indx].pix->n + 1];
         int g = tab_cache[cache_indx].pix->samples[x * tab_cache[cache_indx].pix->n + y * tab_cache[cache_indx].pix->w * tab_cache[cache_indx].pix->n + 2];
         int b = tab_cache[cache_indx].pix->samples[x * tab_cache[cache_indx].pix->n + y * tab_cache[cache_indx].pix->w * tab_cache[cache_indx].pix->n + 3];
         *(ptr+0) = r;
         *(ptr+1) = g;
         *(ptr+2) = b;
         *(ptr+3) = a;
         ptr+=4;
      }
}

void load_page(int cache_indx)
{
  tab_cache[ cache_indx].pageobj = pdf_getpageobject(srcpages, tab_cache[cache_indx].num_page - 1);
  error = pdf_loadpage(&tab_cache[ cache_indx].drawpage, src, tab_cache[ cache_indx].pageobj);
   if (error)
     die(error);
  zoom_page(cache_indx);
}

void load_what_we_need(int num_page,float zoom)
{
  TABCACHE     tab_needed[3];
  int          i;
  int          j;

/*                 dimension 1 is the current page dimension 0 is page minus 1 and dimension 2 is page plus 1                     */

  tab_needed[1].zoom     = zoom;
  tab_needed[1].num_page = num_page;
  tab_needed[1].pix      = nil;
  tab_needed[1].pageobj  = nil;
  tab_needed[1].drawpage = nil;

  if (num_page+1 > pdf_getpagecount( srcpages ))
    num_page = 0;

  tab_needed[2].zoom     = zoom;
  tab_needed[2].num_page = num_page+1;
  tab_needed[2].pix      = nil;
  tab_needed[2].pageobj  = nil;
  tab_needed[2].drawpage = nil;
  
  if (num_page-1 == 0)
      num_page = pdf_getpagecount( srcpages )+1;

  tab_needed[0].zoom     = zoom;
  tab_needed[0].num_page = num_page-1;
  tab_needed[0].pix      = nil;
  tab_needed[0].pageobj  = nil;
  tab_needed[0].drawpage = nil;
 
  triLogError("What we need is\n");
  for(i=0;i<3;i++)
  {
    triLogError("tab[%d] : num_page = %d zoom = %f\n",i,tab_needed[i].num_page,tab_needed[i].zoom);
  }


  /*                           reorganize tab_cache drop page or load page that we need                                          */
  /*                           find if page is in cache                                                                          */
  
  cachesrv_status = 'W';                            /* cache srv temporary unavailable */
  for(i=0;i<3;i++)
  {
    for(j=0;j<3;j++)
      if (tab_needed[i].num_page == tab_cache[j].num_page)
        {
          tab_needed[i].zoom     = tab_cache[j].zoom;
          tab_needed[i].pix      = tab_cache[j].pix;
          tab_needed[i].pageobj  = tab_cache[j].pageobj;
          tab_needed[i].drawpage = tab_cache[j].drawpage;
          
          tab_cache[j].pix       = nil;            /* don't remove is is needed */
        }
  }

  triLogError("What we must remove from cache\n");

  for(i=0;i<3;i++)
  {
    if (tab_cache[i].pix != nil)
      {
        triLogError("tab[%d] : num_page = %d zoom = %f\n",i,tab_cache[i].num_page,tab_cache[i].zoom);
        fz_droppixmap(tab_cache[i].pix);
        fz_dropobj( tab_cache[i].pageobj);
        pdf_droppage( tab_cache[i].drawpage );
      }
  }
  triLogError("********************************\n");
  
  for(i=0;i<3;i++)
  {
    tab_cache[i].zoom        = tab_needed[i].zoom;
    tab_cache[i].num_page    = tab_needed[i].num_page;
    tab_cache[i].pix         = tab_needed[i].pix;
    tab_cache[i].pageobj     = tab_needed[i].pageobj;
    tab_cache[i].drawpage    = tab_needed[i].drawpage;
  }
              
  if (tab_cache[1].pix == nil)
     load_page( 1 );
  else {
          triLogError("page : %d already in cache \n",tab_cache[1].num_page);
       }
  current_pix = tab_cache[1].pix;

  cachesrv_status = 'B';    

  if (tab_cache[2].pix == nil)
     load_page( 2 );

  if (tab_cache[0].pix == nil)
     load_page(  0 ); 

  cachesrv_status = 'R';                                     /* the cache server is loader with 3 images */
  triLogError("cache server is ready 3 images loaded\n");

}
int cache_srv(SceSize args, void *argp)
{
   MSG         *msg;
   int         rc;
   int         i;

   triLogError("Cache server started\n");

   cachesrv_status = 'W';              /* Wait no image ready in cache yet */
   for(i=0;i<3;i++)
   {
     tab_cache[i].zoom     = 0;
     tab_cache[i].num_page = 0;
     tab_cache[i].pix      = nil;
     tab_cache[i].pageobj  = nil;
     tab_cache[i].drawpage = nil;
   }
   error = fz_newrenderer(&drawgc, pdf_devicergb,0,1024*512);
   if (error)
     die(error);
  
   error = pdf_newxref(&src);
   if (error)
     die(error);
   
   error = pdf_loadxref(src, src_pdf);
   if (error)
   {
     fz_printerror(error);
     fz_warn("trying to repair");
     error = pdf_repairxref(src, src_pdf);
     if (error)
	die(error);
    }

   error = pdf_decryptxref(src);
   if (error)
     die(error);

   if (src->crypt)
   {
    // error = pdf_setpassword(src->crypt, password);
    // if (error)
    //	die(error);
      
    }
 
   error = pdf_loadpagetree(&srcpages, src);
   if (error)
     die(error);
  
   rc = sceKernelReceiveMbx(msgbox, (void *)&msg, NULL);
   if (rc < 0)
     {
       die( fz_throw("sceKernelReceiveMBX rc = %d",rc) );
     }
   while (isrunning)
   {
     triLogError("command : %c page : %d zoom : %f\n",msg->command,msg->num_page,msg->zoom);
     switch(msg->command)
     {
       case 'Z':
                while( cachesrv_status == 'W') ;  
                cachesrv_status = 'W';
                tab_cache[1].zoom = msg->zoom; 
                fz_droppixmap(tab_cache[1].pix);
                zoom_page(1);
                current_pix = tab_cache[1].pix;
                cachesrv_status = 'B';
                tab_cache[2].zoom = msg->zoom;
                zoom_page(2);
                tab_cache[0].zoom = msg->zoom;
                zoom_page(0);
                cachesrv_status = 'R';
                break;
       case 'L':load_what_we_need(msg->num_page,msg->zoom);
                break;
     }
     rc = sceKernelReceiveMbx(msgbox,(void *) &msg, NULL);
     if (error < 0)
      {
        die( fz_throw("sceKernelReceiveMBX rc = %d",rc) );
      }
   }
   return 0;
}



