#include "psp_pdf.h"

triImage*    pix_bitmap = nil;

static triS32 next_pow2(triU32 w)
{
  w -= 1;		// To not change already power of 2 values
  w |= (w >> 1);
  w |= (w >> 2);
  w |= (w >> 4);
  w |= (w >> 8);
  w |= (w >> 16);
  return(w+1);
}

static triS32 next_mul8(triU32 w)
{
  return((w+7)&~0x7);
}

void psppdf_init(PARAM *app)
{
  fz_error *error;

  error = fz_newrenderer(&app->rast, pdf_devicergb, 0, 1024 * 512);
  if (error)
    psppdf_error(error);
}

fz_matrix psppdf_viewctm(PARAM *app)
{
  fz_matrix ctm;
  ctm = fz_identity();
  ctm = fz_concat(ctm, fz_translate(0, -app->page->mediabox.y1));
  ctm = fz_concat(ctm, fz_scale(app->zoom, -app->zoom));
  ctm = fz_concat(ctm, fz_rotate(app->rotate + app->page->rotate));

  return ctm;
}

void psppdf_showpage(PARAM *app, int loadpage, int drawpage)
{
  fz_error *error;
  fz_matrix ctm;
  fz_rect bbox;
  fz_obj *obj;

  if (loadpage)
  {
    if (app->page)
      pdf_droppage(app->page);

    app->page = nil;
    obj = pdf_getpageobject(app->pages, app->pageno - 1);

    error = pdf_loadpage(&app->page, app->xref, obj);
    if (error)
      psppdf_error(error);
  }

  if (drawpage)
  {
    unsigned char *ptr;
    int x;
    int y;

    if (app->image)
       fz_droppixmap(app->image);
    app->image = nil;

    ctm = psppdf_viewctm(app);
    bbox = fz_transformaabb(ctm, app->page->mediabox);

    error = fz_rendertree(&app->image, app->rast, app->page->tree,ctm, fz_roundrect(bbox), 1);
    if (error)
      psppdf_error(error);

    ptr = app->image->samples;

    for(y=0; y< app->image->h ;y++)
      for(x=0; x < app->image->w; x++)
       {
         int a = app->image->samples[x * app->image->n + y * app->image->w * app->image->n + 0];
         int r = app->image->samples[x * app->image->n + y * app->image->w * app->image->n + 1];
         int g = app->image->samples[x * app->image->n + y * app->image->w * app->image->n + 2];
         int b = app->image->samples[x * app->image->n + y * app->image->w * app->image->n + 3];
         *(ptr+0) = r;
         *(ptr+1) = g;
         *(ptr+2) = b;
         *(ptr+3) = a;
         ptr+=4;
      } 
  }
}

void psppdf_open(PARAM *app)
{
  fz_error *error;
  fz_obj *obj;
  char *password = "";


//	 Open PDF and load xref table
	 
   error = pdf_newxref(&app->xref);
   if (error)
     psppdf_error(error);

   error = pdf_loadxref(app->xref, app->filename);
   if (error)
   {
     if (!strncmp(error->msg, "ioerror", 7))
	psppdf_error(error);
     psppdf_warn(fz_throw("There was a problem with file \"%s\".\n"
			  "It may be corrupted, or generated by broken software.\n\n"
			  "%s\n\nTrying to continue anyway...",
			  app->filename, error->msg));
     error = pdf_repairxref(app->xref, app->filename);
     if (error)
	psppdf_error(error);
   }
	
//	 * Handle encrypted PDF files
	 

   error = pdf_decryptxref(app->xref);
   if (error)
     psppdf_error(error);

   if (app->xref->crypt)
   {
     triLogError("PDF crypted\n");
     int okay = pdf_setpassword(app->xref->crypt, password);
     while (!okay)
     {
	// password = winpassword(app, filename);
	if (!password)
	   exit(1);
	okay = pdf_setpassword(app->xref->crypt, password);
	if (!okay)
	psppdf_warn(fz_throw("Invalid password."));
     }
   }

	
//	 * Load page tree
	 
   error = pdf_loadpagetree(&app->pages, app->xref);
   if (error)
      psppdf_error(error);
	
    app->doctitle = app->filename;
   
    if (app->xref->info)
    {
	obj = fz_dictgets(app->xref->info, "Title");
	if (obj)
	{
	  error = pdf_toutf8(&app->doctitle, obj);
	  if (error)
	     psppdf_error(error);
	}
    }
	/*
	 * Start at first page
	 */

	if (app->pageno < 1)
		app->pageno = 1;
	if (app->zoom <= 0.0)
		app->zoom = 1.4;
	app->rotate = 0;
	psppdf_showpage(app, 1, 1);
}

void psppdf_close(PARAM *app)
{
	if (app->pages)
		pdf_droppagetree(app->pages);
	app->pages = nil;

	if (app->page)
		pdf_droppage(app->page);
	app->page = nil;

	if (app->image)
		fz_droppixmap(app->image);
	app->image = nil;

	pdf_closexref(app->xref);
	app->xref = nil;
}

void display_page( PARAM *app,int pos_x,int pos_y)
{
   int y;
   unsigned char *ptr_dest;
	
   ptr_dest = pix_bitmap->data; 
   
 for(y=0;y<272;y++)
 {
   if (y > app->image->h)
   {
      memset(ptr_dest,128,480*4);
   }
   else {
          if (app->image->w < 480)
            {
              memset(ptr_dest,128,480*4);
              memcpy(ptr_dest,
                      &app->image->samples[ (pos_y+y)*app->image->n*app->image->w+pos_x*4 ],
                      app->image->w*4);
            }

          else memcpy(ptr_dest,
                      &app->image->samples[ (pos_y+y)*app->image->n*app->image->w+pos_x*4 ],
                      480*4);
        }
   ptr_dest = ptr_dest+((pix_bitmap->stride)*(pix_bitmap->bits>>3));
 }    
// triImageToVRAM( pix_bitmap );
 triDrawImage2(0.0f,0.0f,pix_bitmap);
 if (app->osd_timeout > 0)
 {
   triDrawRect( 32, 32, 140,32,BLACK);
   triFontActivate(impact16);
   triFontPrintf(impact16, 40, 40, WHITE, " Zoom : %1.1f",app->zoom);  // x,y
   app->osd_timeout--;
 }
  
 triSwapbuffers();
}

void render_pdf(void)
{
  int     x                 = 0;
  int     y                 = 0;
  triVec2 *pStick           = nil;
 
//
//  Allocate a bitmap 480x272 in RGBA format
//
  pix_bitmap = triMalloc(sizeof(triImage));

  pix_bitmap->level      = 0; 
  pix_bitmap->levels     = 0;
  pix_bitmap->swizzled   = 0;
  pix_bitmap->palette    = 0;
  pix_bitmap->level      = 0;
  pix_bitmap->palette    = 0;
  pix_bitmap->bits       = 32;
  pix_bitmap->format     = 3;
  pix_bitmap->width      = 480;
  pix_bitmap->height     = 272;
  pix_bitmap->tex_height = next_pow2(272);
  pix_bitmap->stride     = next_pow2(480) ;
  pix_bitmap->size       = pix_bitmap->stride*next_mul8(272)*(pix_bitmap->bits>>3);
  pix_bitmap->data       = triMalloc(pix_bitmap->size);
  memset(pix_bitmap->data,0xff,pix_bitmap->size); 

//
//  Init PDF Lib
//
  psppdf_init(&app);
//
//  Load PDF
//
  psppdf_open(&app);

  triClear( 0x000000 );	  
  triSwapbuffers(); 

  x        = 0;
  y        = 0;

  app.osd_timeout = 300;
  while ( isrunning )
  { 

    display_page( &app,x,y);

    triInputUpdate ();
    pStick	= triInputGetStick ();

    if (triInputHeld (PSP_CTRL_CROSS))
    {
       isrunning = 0;
    }

    if (triInputHeld (PSP_CTRL_LTRIGGER) )
    {
      if (app.zoom > 0.1f)
      {
         app.osd_timeout = 300;
         app.zoom -= 0.1f ;
         psppdf_showpage(&app, 0, 1);
      }
    }

    if (triInputHeld (PSP_CTRL_RTRIGGER) )
    {
      if (app.zoom < 2.1f)
        {
          app.osd_timeout = 300;
          app.zoom += 0.1f;
          psppdf_showpage(&app, 0, 1);
          triLogError("zoom = %f\n",app.zoom);
        }
    }

    if (triInputHeld (PSP_CTRL_LEFT) )
    {
      if (app.pageno > 0)
        app.pageno--;
    }
      
    if (triInputHeld (PSP_CTRL_RIGHT) )
    { 
      if (app.pageno < pdf_getpagecount(app.pages))
        app.pageno++;
    }
     
    if (pStick->x < 0)
      {
        x-=16;
        if (x<0)
          x = 0;
      }
    if (pStick->x > 0)
      {
        x+=16;
        if (x> app.image->w-480)
          x-=16;        
      }
    if (pStick->y < 0)
      {
        y-=8;
        if (y<0)
          y+=8;
      }
    if (pStick->y > 0)
      {
        y+=8;
        if (y>=app.image->h-272)
          y -=8;
      }
  }
  triFree( pix_bitmap->data );
  triFree( pix_bitmap);
}
