#include "psp_pdf.h"

fz_renderer  *drawgc    = nil;
pdf_xref     *src       = nil;
pdf_pagetree *srcpages  = nil;
pdf_page     *drawpage  = nil;
int          dieonerror = 1;
fz_pixmap    *pix;
triImage*    pix_bitmap;


void die(fz_error *eo)
{
  triLogError("+ %s:%d: %s(): %s\n", eo->file, eo->line, eo->func, eo->msg);
}

void fz_printerror(fz_error *eo)
{
   triLogError("+ %s:%d: %s(): %s\n", eo->file, eo->line, eo->func, eo->msg);
}

void printerror(fz_error *eo)
{
   fz_printerror(eo);
}

void maybedie(fz_error* eo)
{
  printerror(eo);
  if (dieonerror)
    sceKernelExitGame();
}

static triS32 next_pow2(triU32 w)
{
	w -= 1;		// To not change already power of 2 values
	w |= (w >> 1);
	w |= (w >> 2);
	w |= (w >> 4);
	w |= (w >> 8);
	w |= (w >> 16);
	return(w+1);
}

static triS32 next_mul8(triU32 w)
{
	return((w+7)&~0x7);
}

void wait_psp(int page,float drawzoom)
{
  char texte[255];

  triClear( 0x000000 );
  triFontActivate(impact16);
  sprintf(texte,"Please Wait");
  triFontPrint(impact16, 150, 110, WHITE, texte); 
  sprintf(texte,"render page : %d",page);
  triFontPrint(impact16, 150, 140, WHITE, texte); 
  sprintf(texte,"zoom = %f",drawzoom);
  triFontPrint(impact16, 150, 170, WHITE, texte);   
  triSwapbuffers();  
}

void close_pdf()
{
  if (srcpages)
  {
    pdf_droppagetree(srcpages);
    srcpages = nil;
  }

  if (src)
  {
    if (src->store)
    {
      pdf_dropstore(src->store);
      src->store = nil;
    }
    pdf_closexref(src);
    src = nil;
  }
}

void load_pdf(char *filename,char *password)
{
   fz_error *error;

   close_pdf();
   error = pdf_newxref(&src);
   if (error)
     die(error);

   error = pdf_loadxref(src, filename);
   if (error)
   {
     fz_printerror(error);
     fz_warn("trying to repair");
     error = pdf_repairxref(src, filename);
     if (error)
	die(error);
    }

   error = pdf_decryptxref(src);
   if (error)
     die(error);

   if (src->crypt)
   {
     error = pdf_setpassword(src->crypt, password);
     if (error)
	die(error);
    }

  
     error = pdf_loadpagetree(&srcpages, src);
     if (error)
	die(error);
}
fz_error *load_page(int pagenum)
{
   fz_obj *pageobj;
   fz_error *error;
   
   pageobj = pdf_getpageobject(srcpages, pagenum - 1);
   drawpage = nil;
   error = pdf_loadpage(&drawpage, src, pageobj);
   if (error)
     {
       maybedie(error);
       return error;
     }
   return nil;
}

fz_error *create_pix(float zoom)
{

   fz_matrix ctm;
   fz_irect bbox;
   int w;
   int h;
   int bh;
   int drawrotate         = 0;
   int drawbands          = 1;
   fz_error *error;
   int x;
   int y;
   unsigned char *ptr;

   ctm = fz_identity();
   ctm = fz_concat(ctm, fz_translate(0, -drawpage->mediabox.y1));
   ctm = fz_concat(ctm, fz_scale(zoom, -zoom));
   ctm = fz_concat(ctm, fz_rotate(drawrotate + drawpage->rotate));

   bbox = fz_roundrect(fz_transformaabb(ctm, drawpage->mediabox));
   w = bbox.x1 - bbox.x0;
   h = bbox.y1 - bbox.y0;
   bh = h / drawbands;
   error = fz_newpixmap(&pix, bbox.x0, bbox.y0, w, bh, 4);
   if (error)
      {
         maybedie(error);
	return error;
      }
   memset(pix->samples, 0xff, pix->h * pix->w * pix->n);
   error = fz_rendertreeover(drawgc, pix, drawpage->tree, ctm);
   if (error)
    { 
	maybedie(error);
	return error;
    }
//
//             Sumatra library use ARGB buffers psp use RGBA so we transform it
//
    ptr = pix->samples;

    for(y=0; y<pix->h ;y++)
      for(x=0; x <pix->w; x++)
       {
         int a = pix->samples[x * pix->n + y * pix->w * pix->n + 0];
         int r = pix->samples[x * pix->n + y * pix->w * pix->n + 1];
         int g = pix->samples[x * pix->n + y * pix->w * pix->n + 2];
         int b = pix->samples[x * pix->n + y * pix->w * pix->n + 3];
         *(ptr+0) = r;
         *(ptr+1) = g;
         *(ptr+2) = b;
         *(ptr+3) = a;
         ptr+=4;
      }
  return nil;
}

void render_pix(int pos_x,int pos_y)
{
   int y;
   unsigned char *ptr_dest;
   char texte[256];
	
   ptr_dest = pix_bitmap->data;
  
   for(y=0;y<272;y++)
   {
    
        memcpy(ptr_dest,&pix->samples[ (pos_y+y)*pix->n*pix->w+pos_x*4 ],480*4);
//      memcpy(ptr_dest,&pix->samples[pos_x+pos_y*pix->w * pix->n+y*pix->w * pix->n],480*4);
     ptr_dest = ptr_dest+((pix_bitmap->stride)*(pix_bitmap->bits>>3));
   }
  triImageToVRAM( pix_bitmap );
  triDrawImage2(0.0f,0.0f,pix_bitmap);
 
   // triDrawSprite( 0.f, 0.f, 0.f, 0.f, pix_bitmap );
  
   
   triSwapbuffers(); 

}
void render_pdf(void)
{
  triVec2 *pStick;
  fz_error *error;
  int page_courante;
  int old_page;
  float    zoom;
  int     x;
  int     y;

//
//  Allocate a bitmap 480x272 in RGBA format
//
  pix_bitmap = triMalloc(sizeof(triImage));

  pix_bitmap->level      = 0; 
  pix_bitmap->levels     = 0;
  pix_bitmap->swizzled   = 0;
  pix_bitmap->palette    = 0;
  pix_bitmap->level      = 0;
  pix_bitmap->palette    = 0;
  pix_bitmap->bits       = 32;
  pix_bitmap->format     = 3;
  pix_bitmap->width      = 480;
  pix_bitmap->height     = 272;
  pix_bitmap->tex_height = next_pow2(272);
  pix_bitmap->stride     = next_pow2(480) ;
  pix_bitmap->size       = pix_bitmap->stride*next_mul8(272)*(pix_bitmap->bits>>3);
  pix_bitmap->data       = triMalloc(pix_bitmap->size);
  memset(pix_bitmap->data,0xff,pix_bitmap->size); 
  /* triLogError("w = %d h = %d texth = %d stride = %d size = %d\n",pix_bitmap->width,
                                                                 pix_bitmap->height,
                                                                 pix_bitmap->tex_height,
                                                                 pix_bitmap->stride,
                                                                 pix_bitmap->size); */
    
  triClear( 0x000000 );	
  
  triSwapbuffers(); 

  error = fz_newrenderer(&drawgc, pdf_devicergb,0,1024*512);
  if (error)
     die(error);

  load_pdf( src_pdf, "" );
  page_courante = 1;
  old_page      = -1;
  zoom = 1.4f;
  x = 0;
  y = 0;
  while ( isrunning )
  {
    if (old_page != page_courante)
      {
         wait_psp(page_courante,zoom);
         old_page = page_courante;
         triSpriteMode( 480, 272, 0 );
         
         triEnable(TRI_PSEUDO_FSAA);
         triEnable(TRI_VBLANK); 
         error = load_page( page_courante );
         if (error)
         {
           isrunning = 0;
           return;
         }
         error = create_pix( zoom );
         if (error)
         {
           isrunning = 0;
           return;
         }
      }     
    render_pix( x , y );
    triInputUpdate ();
    pStick	= triInputGetStick ();

    if (triInputHeld (PSP_CTRL_CROSS))
    {
       isrunning = 0;
       return;
    }
   if (triInputHeld (PSP_CTRL_LTRIGGER) )
     {
       zoom -= 0.1 ;
       wait_psp(page_courante,zoom); 
       fz_droppixmap(pix);
       error = create_pix( zoom );
         if (error)
         {
           isrunning = 0;
           return;
         }
      
     }
   if (triInputHeld (PSP_CTRL_RTRIGGER) )
     {
       zoom += 0.1 ;
       fz_droppixmap(pix);
      wait_psp(page_courante,zoom); 

       error = create_pix( zoom );
       if (error)
         {
           isrunning = 0;
           return;
         }
     }
   if (pStick->x < 0)
        {
          x-=16;
          if (x<0)
            x = 0;
        }
   if (pStick->x > 0)
        {
          x+=16;
          if (x> pix->w-480)
             x-=16;
         
        }
   if (pStick->y < 0)
       {
          y-=8;
             if (y<0)
               y+=8;
       }
      if (pStick->y > 0)
       {
          y+=8;
             if (y>=pix->h-272)
               y -=8;
       }
 
   
  }
  
}
